package org.ioc.commons.impl.gwt.integration.client.dataaccess.requestfactory.rebind.internal;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.ioc.commons.flowcontrol.common.BindRegistration;
import org.ioc.commons.flowcontrol.operationmanager.IsOperation;
import org.ioc.commons.flowcontrol.operationmanager.OperationManagerTrigger;
import org.ioc.commons.integration.dataaccess.dao.EntityDao;
import org.ioc.commons.integration.dataaccess.dao.IsDaoFactory;
import org.ioc.commons.integration.dataaccess.dao.SubscriptionChangeHandler;
import org.ioc.commons.integration.dataaccess.dao.SubscriptionChangeHandler.ChangeType;
import org.ioc.commons.integration.dataaccess.dao.SubscriptionChangesBinder;
import org.ioc.commons.utils.FormatterLogger;

import com.google.web.bindery.event.shared.EventBus;
import com.google.web.bindery.event.shared.HandlerRegistration;
import com.google.web.bindery.requestfactory.shared.EntityProxy;
import com.google.web.bindery.requestfactory.shared.EntityProxyChange;
import com.google.web.bindery.requestfactory.shared.RequestContext;
import com.google.web.bindery.requestfactory.shared.RequestFactory;
import com.google.web.bindery.requestfactory.shared.ServerFailure;

public class RFSubscriptionBinder<P extends EntityProxy, I, D extends IsDaoFactory> implements
		SubscriptionChangesBinder<P, I> {

	private static FormatterLogger logger = FormatterLogger.getLogger(RFSubscriptionBinder.class);

	private Map<SubscriptionChangeHandler<I>, BindRegistration> onceChangeHandlers = new HashMap<SubscriptionChangeHandler<I>, BindRegistration>();

	private P entityProxy;
	private Class<P> entityProxyClass;
	private EventBus gwtEventBus;
	private RFBaseEntityDao<P, ? extends RequestContext, I, D, ? extends RequestFactory> entityDao;

	private static class GwtSubscriptorBindRegistration implements BindRegistration {

		private HandlerRegistration subscriptorHandlerRegistration;

		public GwtSubscriptorBindRegistration(HandlerRegistration subscriptorHandlerRegistration) {
			this.subscriptorHandlerRegistration = subscriptorHandlerRegistration;
		}

		@Override
		public void unbind() {
			this.subscriptorHandlerRegistration.removeHandler();
		}

	}

	@SuppressWarnings("unchecked")
	public RFSubscriptionBinder(Class<P> entityProxyClass, EventBus gwtEventBus, EntityDao<P, I, D> entityDao) {
		this.entityProxyClass = entityProxyClass;
		this.gwtEventBus = gwtEventBus;
		this.entityDao = (RFBaseEntityDao<P, ? extends RequestContext, I, D, ? extends RequestFactory>) entityDao;
	}

	@Override
	public BindRegistration bindChanges(final SubscriptionChangeHandler<I> handler) {
		HandlerRegistration subscriptorHandlerRegistration = EntityProxyChange.<P> registerForProxyType(gwtEventBus,
				entityProxyClass, new EntityProxyChange.Handler<P>() {

					@Override
					public void onProxyChange(EntityProxyChange<P> gwtEvent) {
						switch (gwtEvent.getWriteOperation()) {
						case UPDATE:
							if (entityProxy == null || gwtEvent.getProxyId().equals(entityProxy.stableId())) {
								handler.handleSubscription(gwtEvent.getSource(),
										entityDao.getId(gwtEvent.getProxyId()), ChangeType.Updated);

								removeIfOnce(handler);
							}
							break;
						case DELETE:
							if (entityProxy == null || gwtEvent.getProxyId().equals(entityProxy.stableId())) {
								handler.handleSubscription(gwtEvent.getSource(),
										entityDao.getId(gwtEvent.getProxyId()), ChangeType.Deleted);

								removeIfOnce(handler);
							}
							break;
						case PERSIST:
							if (entityProxy == null || gwtEvent.getProxyId().equals(entityProxy.stableId())) {
								handler.handleSubscription(gwtEvent.getSource(),
										entityDao.getId(gwtEvent.getProxyId()), ChangeType.Persisted);
								removeIfOnce(handler);
							}
							break;
						}
					}
				});

		return new GwtSubscriptorBindRegistration(subscriptorHandlerRegistration);
	}

	@Override
	public BindRegistration bindChangesOnce(SubscriptionChangeHandler<I> handler) {
		BindRegistration registration = bindChanges(handler);

		this.onceChangeHandlers.put(handler, registration);

		return registration;
	}

	@Override
	public BindRegistration updateList(final List<P> collection, String... propertyRefs) {
		return updateList(null, null, collection, propertyRefs);
	}

	@Override
	public BindRegistration updateListOnce(List<P> collection, String... propertyRefs) {
		return updateListOnce(null, null, collection, propertyRefs);
	}

	@Override
	public <O extends Enum<? extends IsOperation>> BindRegistration updateList(O operation,
			OperationManagerTrigger<O> trigger, List<P> collection, String... propertyRefs) {
		return updateList(operation, trigger, collection, false, propertyRefs);
	}

	@Override
	public <O extends Enum<? extends IsOperation>> BindRegistration updateListOnce(O operation,
			OperationManagerTrigger<O> trigger, List<P> collection, String... propertyRefs) {
		return updateList(operation, trigger, collection, true, propertyRefs);
	}

	protected <O extends Enum<? extends IsOperation>> BindRegistration updateList(final O operation,
			final OperationManagerTrigger<O> trigger, final List<P> collection, boolean once,
			final String... propertyRefs) {
		SubscriptionChangeHandler<I> handler = new SubscriptionChangeHandler<I>() {

			@Override
			public void handleSubscription(Object sender, final I id, SubscriptionChangeHandler.ChangeType changeType) {
				if (collection == null || collection.isEmpty()) {
					return;
				}

				/**
				 * Searching if the list contains the element.
				 */
				if (changeType == ChangeType.Persisted) {
					if (trigger != null && operation != null) {
						trigger.operationBeginning(operation);
					}
					entityDao.getDataFactory().find(((RFBaseIdImpl) id).getEntityProxyId()).with(propertyRefs)
							.to(new RFDaoReceiver<EntityProxy>() {

								@Override
								public void onEnd(boolean success) {
									super.onEnd(success);
									if (trigger != null && operation != null) {
										trigger.operationFinished(operation, success);
									}
								}

								@Override
								public void success(EntityProxy response) {
									super.success(response);
									@SuppressWarnings("unchecked")
									P element = (P) response;
									collection.add(element);
									removeIfOnz();
								}

								@Override
								public void failure(ServerFailure error) {
									super.failure(error);
									logger.error(error.getMessage());
								}
							}).fire();
				} else {
					for (int i = collection.size() - 1; i >= 0; i--) {
						P element = collection.get(i);
						if (entityDao.getId(element).equals(id)) {
							switch (changeType) {
							case Persisted:
								// Tratado arriba
								break;
							case Updated:

								if (trigger != null && operation != null) {
									trigger.operationBeginning(operation);
								}

								/**
								 * Finding the element
								 */
								entityDao.getDataFactory().find(((RFBaseIdImpl) id).getEntityProxyId())
										.with(propertyRefs).with(propertyRefs).to(new RFDaoReceiver<EntityProxy>() {

											public void onEnd(boolean success) {
												if (trigger != null && operation != null) {
													trigger.operationFinished(operation, success);
												}
											};

											@Override
											public void success(EntityProxy response) {
												super.success(response);
												/**
												 * Searching the element again
												 * in order to avoid possible
												 * modifications and order
												 * changes.
												 */
												for (int i = collection.size() - 1; i >= 0; i--) {
													P element = collection.get(i);
													if (entityDao.getId(element).equals(id)) {
														@SuppressWarnings("unchecked")
														P newElement = (P) response;
														collection.set(i, newElement);
														removeIfOnz();
														return;
													}
												}
											}

											@Override
											public void failure(ServerFailure error) {
												super.failure(error);
												logger.error(error.getMessage());
											}
										}).fire();
								break;
							case Deleted:
								collection.remove(i);
								removeIfOnz();
								break;
							}

							return;
						}
					}
				}

			}

			private void removeIfOnz() {
				removeIfOnce(this);
			}
		};
		BindRegistration registration = bindChanges(handler);

		if (once) {
			this.onceChangeHandlers.put(handler, registration);
		}

		return registration;
	}

	public P getEntityProxy() {
		return entityProxy;
	}

	public void setEntityProxy(P entityProxy) {
		this.entityProxy = entityProxy;
	}

	protected void removeIfOnce(final SubscriptionChangeHandler<I> handler) {
		if (!onceChangeHandlers.isEmpty()) {
			BindRegistration registration = onceChangeHandlers.remove(handler);
			if (registration != null) {
				registration.unbind();
			}
		}
	}

}
